-- scope stuff
gearquipper = gearquipper or {};
local c = gearquipper;

local switchQueue, switching = {};
local inCombat = false;

local lastRealmname, lastCharname;
local lastMountState, lastStealthState, lastPvPState, lastAfkState, lastPartyState, lastRaidState, lastSubmergedState;
local lastRaidRoster, lastMapId;

function c:GetCharName()
	return UnitName("player");
end

function c:GetRealmName()
	return GetRealmName();
end

function c:NotifyInChat(msg)
	if GQ_OPTIONS[c.OPT_HIGHLIGHTCHANGES] then
		c:Println(msg);
	end
end

local function HighlightButton(button)
	if GQ_OPTIONS[c.OPT_HIGHLIGHTCHANGES] and button then
		ActionButton_ShowOverlayGlow(button);
		C_Timer.After(1, function () ActionButton_HideOverlayGlow(button); end);
	end
end

function c:HighlightInventorySlot(slotId)
	if c:GetSlotInfo()[slotId] then
		HighlightButton(_G["Character" .. c:GetSlotInfo()[slotId]]);
	end
end

function c:HighlightActionSlot(slotId)
	HighlightButton(c:GetActionButton(slotId));
end

function c:IsSwitching()
	return switching;
end

-- command handling
SLASH_GEARQUIPPER1 = "/gq";
local CMD_RESET = "reset";
local CMD_SWITCH = "switch";

SlashCmdList["GEARQUIPPER"] = function(msg)
	if msg then
		if c.initFinished then
			msg = c:Trim(msg);

			if msg == CMD_RESET then
				c:ShowResetDialog();
				return;
			elseif c:StartsWith(msg, CMD_SWITCH, true) then
				msg = c:Trim(msg:gsub(CMD_SWITCH, ""));
				local existingSets, setNames = c:LoadSetNames(), {};
				if msg ~= "" then
					for v in string.gmatch(msg, "[^ ]+") do
						if string.upper(v) == c.KEYWORD_PREVIOUS then
							table.insert(setNames, c.KEYWORD_PREVIOUS);
						else
							local existingSetName = c:TableContains(existingSets, v, true);
							if existingSetName then
								table.insert(setNames, existingSetName);
							else
								c:Println(c:GetText("Macro warning: There is no set named \"%s\".", v));
							end
						end
					end
				else
					setNames = c:Deepcopy(existingSets);
				end

				local noSets = table.getn(setNames);
				if noSets == 1 then
					c:QueueSwitch(setNames[1]);
				elseif noSets > 1 then
					local currentSet = c:LoadCurrentSetName();
					for k, v in ipairs(setNames) do
						if v == currentSet then
							local index = k + 1;
							if index > noSets then
								index = 1;
							end
							c:QueueSwitch(setNames[index]);
							return;
						end
					end
					c:QueueSwitch(setNames[1]);
				end
				return;
			end

			c:Println(c:GetText("Unknown command. Possible parameters are: /gq ..."));
			c:Println(c:GetText("reset -> Resets the GearQuipper addon (e.g. in case of errors)."));
			c:Println(c:GetText("switch [setname1] [setname2] ... -> Switches between specified sets. For macro use."));
		else
			c:Println(c:GetText("The AddOn has not been initialized due to an error. Please relog and try again."));
		end
	end
end

-- event handling
c.eventFrame = c.eventFrame or CreateFrame("Frame");
c.eventFrame:RegisterEvent("ADDON_LOADED");
c.eventFrame:RegisterEvent("PLAYER_ENTERING_WORLD");

c.eventFrame:RegisterEvent("PLAYER_FLAGS_CHANGED");
c.eventFrame:RegisterEvent("PLAYER_REGEN_ENABLED");
c.eventFrame:RegisterEvent("PLAYER_REGEN_DISABLED");
c.eventFrame:RegisterEvent("SPELL_UPDATE_USABLE");
c.eventFrame:RegisterEvent("MIRROR_TIMER_START");
c.eventFrame:RegisterEvent("MIRROR_TIMER_STOP");
c.eventFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");

c.eventFrame:RegisterEvent("UNIT_INVENTORY_CHANGED");
c.eventFrame:RegisterEvent("PLAYER_EQUIPMENT_CHANGED");
c.eventFrame:RegisterEvent("ACTIONBAR_SLOT_CHANGED");
c.eventFrame:RegisterEvent("PLAYER_LEVEL_UP");

c.eventFrame:RegisterEvent("ITEM_LOCKED");
c.eventFrame:RegisterEvent("ITEM_UNLOCKED");

--c.eventFrame:RegisterEvent("UNIT_SPELLCAST_SENT");
--c.eventFrame:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED");
c.eventFrame:RegisterEvent("UNIT_AURA");

c.eventFrame:RegisterEvent("GROUP_ROSTER_UPDATE");
c.eventFrame:RegisterEvent("RAID_ROSTER_UPDATE");

c.eventFrame:RegisterEvent("ZONE_CHANGED");
c.eventFrame:RegisterEvent("ZONE_CHANGED_INDOORS");
c.eventFrame:RegisterEvent("ZONE_CHANGED_NEW_AREA");

c.eventFrame:HookScript("OnEvent", function(self, event, arg1, arg2, arg3, ...)
	if event then
		if event == "ADDON_LOADED" and arg1 == "GearQuipper" and not c.initFinished then
			c:Init();
		elseif event == "PLAYER_ENTERING_WORLD" then
			local currentRealmName, currentCharName = GetRealmName(), c:GetCharName();
			if not c.currentSession or c.currentSession["realmName"] ~= currentRealmName or c.currentSession["charName"] ~= currentCharName then
				-- create default gear if not existing
				if table.getn(c:LoadSetNames()) == 0 then
					c:SaveDefaultSet();
				end

				-- set initial states
				lastMountState = IsMounted();
				lastStealthState = IsStealthed();
				lastPartyState = IsInGroup();
				lastRaidState = IsInRaid();
				lastPvPState = UnitIsPVP("player");
				lastAfkState = UnitIsAFK("player");
				lastSubmergedState = IsSwimming();
				lastMapId = C_Map.GetBestMapForUnit("player");

				-- set cloak/helmet settings
				c:LoadCloakAndHelmet();

				c.currentSession = { ["realmName"] = currentRealmName, ["charName"] = currentCharName };
			else
				c:NewZone();
			end
		elseif event == "UNIT_INVENTORY_CHANGED" and arg1 == "player" and c:SaveConditionsMet() then
			-- ammo slot workaround
			local currentSetName = c:LoadCurrentSetName();
			if currentSetName and (not c:LoadPartialOption() or c:LoadSlotState(arg1)) and not c:IsSetItem(INVSLOT_AMMO, c:LoadSlot(INVSLOT_AMMO, currentSetName)) then
				c:SaveSlot(INVSLOT_AMMO, currentSetName);
				c:SetSlotInfo(INVSLOT_AMMO, "green");
			end
		elseif event == "PLAYER_EQUIPMENT_CHANGED" and c:SaveConditionsMet() then
			if c:GetSlotInfo()[arg1] and not c:IsSwitching() then
				local currentSetName = c:LoadCurrentSetName();
				if currentSetName and (not c:LoadPartialOption() or c:LoadSlotState(arg1)) then
					if not c:IsSetItem(arg1, c:LoadSlot(arg1, currentSetName)) then
						c:SaveSlot(arg1, currentSetName);
						c:SetSlotInfo(arg1, "green");
					end
				else
					if c:LoadPartialOption() and not c:LoadSlotState(arg1) then
						c:SetSlotInfo(arg1, "white");
					else
						c:SetSlotInfo(arg1, "red");
					end
				end
			end
		elseif event == "ACTIONBAR_SLOT_CHANGED" then
			if not c:IsSwitching() and c:LoadActionSlotsOption() then
				c:SaveActionSlot(arg1);
			end
		elseif event == "PLAYER_LEVEL_UP" then
			c:LevelUp(arg1);
		elseif event == "COMBAT_LOG_EVENT_UNFILTERED" then
			local currentPvPState = UnitIsPVP("player");
			if lastPvPState ~= currentPvPState then
				if currentPvPState then
					c:HandleEvent(c.EVENT_PVP_ENABLE);
				else
					c:HandleEvent(c.EVENT_PVP_DISABLE);
				end
				lastPvPState = currentPvPState;
			end
		elseif event == "ITEM_LOCKED" and not arg2 then
			if not c:IsSwitching() then
				local slotId = tonumber(arg1);
				if slotId >= 0 and slotId <= 19 then
					c:OpenQuickBar(slotId);
				end
			end
		elseif event == "ITEM_UNLOCKED" then
			c:CloseQuickBars();
		elseif event == "UNIT_AURA" then
			C_Timer.After(1, function()
				local currentMountState, currentStealthState = IsMounted(), IsStealthed();
				if currentMountState ~= lastMountState then
					if currentMountState then
						c:HandleEvent(c.EVENT_MOUNT);
					else
						c:HandleEvent(c.EVENT_DISMOUNT);
					end
					lastMountState = currentMountState;
				end
				if currentStealthState ~= lastStealthState then
					if currentStealthState then
						c:HandleEvent(c.EVENT_STEALTH);
					else
						c:HandleEvent(c.EVENT_UNSTEALTH);
					end
					lastStealthState = currentStealthState;
				end
			end);
		elseif event == "PLAYER_REGEN_ENABLED" then
			c:LeaveCombat();
		elseif event == "PLAYER_REGEN_DISABLED" then
			c:EnterCombat();
		elseif event == "PLAYER_FLAGS_CHANGED" then
			local currentAfkState = UnitIsAFK("player");
			if lastAfkState ~= currentAfkState then
				if currentAfkState then
					c:HandleEvent(c.EVENT_AFK_ENABLE);
				else
					c:HandleEvent(c.EVENT_AFK_DISABLE);
				end
				lastAfkState = currentAfkState;
			end
		elseif event == "GROUP_ROSTER_UPDATE" or event == "RAID_ROSTER_UPDATE" then
			local currentPartyState, currentRaidState = IsInGroup(), IsInRaid();
			if lastPartyState ~= currentPartyState then
				if currentPartyState then
					c:HandleEvent(c.EVENT_PARTY_JOIN);
				else
					c:HandleEvent(c.EVENT_PARTY_LEAVE);
				end
				lastPartyState = currentPartyState;
			end

			if lastRaidState ~= currentRaidState then
				if currentRaidState then
					c:HandleEvent(c.EVENT_RAID_JOIN);
				else
					c:HandleEvent(c.EVENT_RAID_LEAVE);
				end
				lastRaidState = currentRaidState;
			end
		elseif c:StartsWith(event, "ZONE_CHANGED") then
			if c.firstZoneEntered then
				c:NewZone();
			else
				c.firstZoneEntered = true;
			end
		elseif event == "SPELL_UPDATE_USABLE" or "MIRROR_TIMER_START" or "MIRROR_TIMER_STOP" then
			local currentSubmergedState = IsSwimming();
			if lastSubmergedState ~= currentSubmergedState then
				if currentSubmergedState then
					c:HandleEvent(c.EVENT_SUBMERGE);
				else
					c:HandleEvent(c.EVENT_EMERGE);
				end
				lastSubmergedState = currentSubmergedState;
			end
		--elseif event == "UNIT_SPELLCAST_SENT" then
		--	c:QueueSwitch("test1")
		end
	end
end);

function c:HandleEvent(eventType, eventSubType)
	if eventType == c.EVENT_MOUNT then
		lastMountState = true;
	elseif eventType == c.EVENT_DISMOUNT then
		lastMountState = false;
	elseif eventType == c.EVENT_STEALTH then
		lastStealthState = true;
	elseif eventType == c.EVENT_UNSTEALTH then
		lastStealthState = false;
	end

	local filter = { [c.FIELD_TYPE] = eventType };
	if eventType == c.EVENT_ZONE_ENTER or eventType == c.EVENT_ZONE_LEAVE then
		filter[c.FIELD_SUBTYPE] = c:GetZoneInfo(eventSubType);
	else

	end
	local setBindings = c:LoadEventBindings(filter);
	if setBindings then
		local currentPvPState = UnitIsPVP("player");
		for index, binding in pairs(setBindings) do
			if (binding[c.FIELD_PVE] and not currentPvPState) or (binding[c.FIELD_PVP] and currentPvPState) then
				c:QueueSwitch(binding[c.FIELD_NAME]);
				return;
			end
		end
	end
end

function c:NewZone()
	local currentMapId = C_Map.GetBestMapForUnit("player");
	if currentMapId ~= lastMapId then
		if lastMapId then
			-- leave bg
			if c:IsZoneBattleground(lastMapId) then
				c:HandleEvent(c.EVENT_BG_LEAVE);
			end

			-- leave zone
			local lastParentMapId = c:GetZoneInfo(lastMapId)["parentMapId"];
			if lastParentMapId and lastParentMapId ~= 947 then
				-- leave parent zone
				c:HandleEvent(c.EVENT_ZONE_LEAVE, lastParentMapId);
			end
			c:HandleEvent(c.EVENT_ZONE_LEAVE, lastMapId);
		end
		lastMapId = currentMapId;

		if currentMapId then
			-- enter bg
			if c:IsZoneBattleground(currentMapId) then
				c:HandleEvent(c.EVENT_BG_ENTER);
			end

			-- enter zone
			local currentParentMapId = c:GetZoneInfo(currentMapId)["parentMapId"];
			if currentParentMapId and currentParentMapId ~= 947 then
				-- enter parent zone
				c:HandleEvent(c.EVENT_ZONE_ENTER, currentParentMapId);
			end
			c:HandleEvent(c.EVENT_ZONE_ENTER, currentMapId);
		end
	end
end

function c:EnterCombat()
	inCombat = true;
	c:LockUI();
end

function c:LeaveCombat()
	inCombat = false;
	c:UnlockUI();
end

function c:IsInCombat()
	return InCombatLockdown() or inCombat;
end

function c:IsOnBattleground()
	local bgPosition = UnitInBattleground("player");
	if not bgPosition then
		return false;
	end
	return bgPosition;
end

function c:FinishSwitch(setName, retry)
	C_Timer.After(1, function()
		if retry and c:IsSetComplete(setName) then
			c:Println(c:GetText("Switch to set \"%s\" complete.", setName));
		end
		switching = nil;

		if table.getn(switchQueue) > 0 then
			local nextSet = table.remove(switchQueue, 1);
			c:SwitchToSet(nextSet);
		else
			c:LoadCloakAndHelmet(setName);			
			if PaperDollFrame:IsVisible() then
				c:SetPaperDollLabelText();
				c:SetSlotInfo();
				c:ShowSlotStateBoxes();
				c:RefreshSetList(setName);
				c:UnlockUI();
				PlaySound(1264);
			end
		end
	end);
end

function c:QueueSwitch(setName)
	if not c:IsSwitching() then
		c:SwitchToSet(setName);
	else
		table.insert(switchQueue, setName);
	end
end

function c:SwitchToSet(setName, retry)
	if (not c:IsSwitching() or retry) and setName then
		if setName == c.KEYWORD_PREVIOUS then
			setName = c:LoadPreviousSetName();
		end

		setName = c:TableContains(c:LoadSetNames(), setName, true); -- case insensitive
		if setName then

			local freeSpace, neededSpace, bagSpaceCache = c:CheckNeccessaryBagSpace(setName);
			if neededSpace > freeSpace then
				c:Println(c:GetText("Not enough bag space. Current space: %s, needed: %s", freeSpace, neededSpace));
				c:RefreshSetList();
			else
				switching = setName;
				c:LockUI(setName);
				PlaySound(1264);

				for _, slotId in ipairs(c:GetSlotSwitchOrder()) do
					local desiredItemString = c:LoadSlot(slotId, setName);

					if not c:IsSetItem(slotId, desiredItemString) and (not c:LoadPartialOption(setName) or c:LoadSlotState(slotId, setName)) then
						if InCombatLockdown() then -- and not (slotId == INVSLOT_MAINHAND or slotId == INVSLOT_OFFHAND or slotId == INVSLOT_RANGED) then
							-- breaks loop if combat starts
							if not retry then
								c:Println(c:GetText("Switching interrupted by combat. It will be re-attempted once combat is over."));
								c:RefreshSetList();
							end
							C_Timer.After(1, function() c:SwitchToSet(setName, true); end);
							return;
						end

						if c:IsEmpty(desiredItemString) then
							-- unequip item
							local itemString = c:GetItemString(slotId);

							ClearCursor();
							PickupInventoryItem(slotId);
							if not c:PutInBag(bagSpaceCache, itemString) then
								c:Println(c:GetText("Not enough bag space. Switching aborted."));
								c:RefreshSetList();
								switching = nil;
								c:UnlockUI();
								return;
							end

							ClearCursor();
						else
							-- equip item
							local itemLink = c:GetItemLink(desiredItemString);
							if itemLink then
								local bagId, bagSlotId = c:FindItemInBags(desiredItemString);
								if bagId then
									c:SaveLastBagLocation(desiredItemString, bagId, bagSlotId);
									if c:IsEmpty(slotId) then
										bagSpaceCache[bagId] = bagSpaceCache[bagId] + 1;
									end
								end

								if slotId == 0 then
									-- ammo workaround
									EquipItemByName(itemLink);
								else
									EquipItemByName(itemLink, slotId);
								end
							end
						end

						if GQ_OPTIONS[c.OPT_HIGHLIGHTCHANGES] then
							c:HighlightInventorySlot(slotId);
						end
					end
				end

				if c:LoadActionSlotsOption(setName) then
					c:LoadActionConfiguration(setName, retry);
				end

				c:SaveCurrentSetName(setName);
				c:FinishSwitch(setName, retry);
			end
		end
	end
end

function c:LoadActionConfiguration(setName, retry)
	local actionSlots = c:LoadActionSlots(setName);
	for slotId, entry in ipairs(actionSlots) do
		if entry and not c:IsSameAction(slotId, entry) then
			if InCombatLockdown() then
				-- breaks loop if combat starts
				if not retry then
					c:Println(c:GetText("Switching interrupted by combat. It will be re-attempted once combat is over."));
				end
				C_Timer.After(1, function() c:SwitchToSet(setName, true); end);
				return;
			end

			ClearCursor();
			PickupAction(slotId);
			ClearCursor();

			if not c:IsEmpty(entry) then
				if entry[c.FIELD_TYPE] == "spell" then
					PickupSpell(entry[c.FIELD_ID]);
				elseif entry[c.FIELD_TYPE] == "item" then
					PickupItem(entry[c.FIELD_ID]);
				elseif entry[c.FIELD_TYPE] == "macro" then
					PickupMacro(entry[c.FIELD_ID]);
				end

				PlaceAction(slotId);
				ClearCursor();
			end

			c:HighlightActionSlot(slotId);
		end
	end

	return true;
end

function c:PutInBag(bagSpaceCache, itemString)
	if itemString then
		local lastBagId = c:LoadLastBagLocation(itemString);
		if lastBagId and bagSpaceCache[lastBagId] > 0 and c:PutInBackpack(bagSpaceCache, lastBagId) then
			return true;
		end
	end

	for index in pairs(bagSpaceCache) do
		if bagSpaceCache[index] > 0 then
			return c:PutInBackpack(bagSpaceCache, index);
		end
	end
end

function c:PutInBackpack(bagSpaceCache, bagId)
	if CursorHasItem() then
		if bagId == 0 then
			PutItemInBackpack();
		else
			PutItemInBag(ContainerIDToInventoryID(bagId));
		end

		if not CursorHasItem() then
			bagSpaceCache[bagId] = bagSpaceCache[bagId] - 1;
			return true;
		end
	end
end